/* file sample.c
   sample OS/2 serial device driver

   Steve Mastrianni
   290 Brookfield Street
   South Windsor, CT 06074
   (203) 644-0723 voice
   (203) 644-4764 fax/data
*/

#include "drvlib.h"
#include "uart.h"
#include "sample.h"

extern void near STRAT();           /* name of strat rout. in DRVSTART      */
extern void near TIMER();           /* timer handler                       */
extern int  near INT_HNDLR();       /* interrupt handler in drvstart        */

/* device header */

DEVICEHDR devhdr = {
        (void far *) 0xFFFFFFFF,     /* link (no more devices)             */
        (DAW_CHR | DAW_OPN | DAW_LEVEL),/* attribute                       */
        (OFF) STRAT,                 /* &strategy                          */
        (OFF) 0,                     /* &IDCroutine                        */
        "DEVICE$ "
        };

/* variables and other stuff */

CHARQUEUE   rx_queue;               /* receiver queue                      */
CHARQUEUE   tx_queue;               /* transmitter queue                   */
FPFUNCTION  DevHlp=0;               /* storage area for DevHlp calls       */
LHANDLE     lock_seg_han;           /* handle for locking appl. segment    */
PHYSADDR    appl_buffer=0;          /* address of caller's buffer          */
PREQPACKET  p=0L;                   /* request packet pointer              */
ERRCODE     err=0;                  /* error return                        */
FARPOINTER  *ptr;                   /* temp far pointer                    */
USHORT      i;                      /* general counter                     */
UARTREGS    uart_regs;              /* uart configuration structure        */
ULONG       WriteID=0L;             /* block ID for write                  */
ULONG       ReadID=0L;              /* block ID for read                   */
PREQPACKET  ThisReadRP=0L;          /* read request packet pointer         */
PREQPACKET  ThisWriteRP=0L;         /* write request packet pointer        */
UCHAR       inchar,outchar;         /* temp for input and output chars     */
USHORT      baud_rate=9600;         /* save the current baud rate          */
USHORT      savepid;                /* PID of caller                       */
UCHAR       opencount;              /* count of DosOpens                   */
ULONG       tickcount;              /* timeout counter                     */
USHORT      com_error_word;         /* status of port                      */
USHORT      port;                   /* temp port variable                  */
QUEUE       rqueue;                 /* return queue status                 */

void near init();
void near enable_write();
void near disable_write();
void near set_dlab();
void near reset_dlab();
void near config_82050();

char     IntFailMsg[]  = "interrupt handler failed to install.\r\n";
char     TimFailMsg[]  = "timer handler failed to install.\r\n";
char     MainMsg[]     = "sample OS/2 Serial Device Driver V1.0 installed.\r\n";

/* Device Initialization Routine */

int Init(PREQPACKET rp, int dev)
{
    register char far *p;

    DevHlp = rp->s.Init.DevHlp;     /* save the DevHlp entry point         */

    /* install timer handler */

    if (SetTimer((PFUNCTION)TIMER))  {
       DosPutMessage(1, 8, devhdr.DHname);
       DosPutMessage(1,strlen(TimFailMsg),TimFailMsg);
       goto fail;
      }

    rx_queue.qsize=QUEUE_SIZE;
    tx_queue.qsize=QUEUE_SIZE;      /* init queue headers                  */
    init();                         /* init the port                       */
    tickcount=MIN_TIMEOUT;          /* set timeout                         */

    /* install the interrupt handler */

    if(SetIRQ(5,(PFUNCTION)INT_HNDLR,0)) {

    /* if we failed, effectively deinstall driver with cs+ds=0 */

       DosPutMessage(1, 8, devhdr.DHname);
       DosPutMessage(1,strlen(IntFailMsg),IntFailMsg);
fail:
       rp->s.InitExit.finalCS = (OFF) 0;
       rp->s.InitExit.finalDS = (OFF) 0;
       return (RPDONE | RPERR | ERROR_BAD_COMMAND);
       }

    /* output initialization message */

    DosPutMessage(1, 8, devhdr.DHname);
    DosPutMessage(1, strlen(MainMsg), MainMsg);

    /* send back our cs and ds end values to os/2 */

    if (SegLimit(HIUSHORT((void far *) Init), &rp->s.InitExit.finalCS) ||
        SegLimit(HIUSHORT((void far *) MainMsg), &rp->s.InitExit.finalDS))
        Crash();
    return(RPDONE);
}

/* common entry point for calls to strat routines */

int main(PREQPACKET rp, int dev )
{
    void far *ptr;
    PLINFOSEG liptr;                /* pointer to local info seg           */
    int i;
    ULONG addr;

    switch(rp->RPcommand)
    {
    case RPINIT:                    /* 0x00                                */

        /* init called by kernel in protected mode */

        return Init(rp,dev);

    case RPOPEN:                    /* 0x0d                                */

        /* get current processes id */

        if (GetDOSVar(2,&ptr))
            return (RPDONE | RPERR | ERROR_BAD_COMMAND);

        /* get process info */

        liptr = *((PLINFOSEG far *) ptr);

        /* if this device never opened, can be opened by any process       */

        if (opencount == 0)         /* first time this device opened       */
        {
            ThisReadRP=0L;          /* clear read and write packet pointers*/
            ThisWriteRP=0L;
            opencount=1;            /* set open counter                    */
            savepid = liptr->pidCurrent; /* save current process id        */
            QueueInit(&rx_queue);   /* init driver queues for this uart    */
            QueueInit(&tx_queue);
        }
        else
            {
            if (savepid != liptr->pidCurrent) /* another proc tried to open */
                return (RPDONE | RPERR | RPBUSY ); /* so return error       */
            ++opencount;            /* bump counter, same pid               */
        }
        return (RPDONE);

    case RPCLOSE:                   /* 0x0e                                */

        /* get process info of caller */

        if (GetDOSVar(2,&ptr))
            return (RPDONE | RPERR | ERROR_BAD_COMMAND); /* no info        */

        /* get process info from os/2 */

        liptr= *((PLINFOSEG far *) ptr); /* ptr to process info seg        */

        /*
        make sure that process attempting to close this device
        one that originally opened it and the device was open in
        first place.
        */

        if (savepid != liptr->pidCurrent || opencount == 0)
            return (RPDONE | RPERR | ERROR_BAD_COMMAND);

        --opencount;                /* close counts down open counter      */
        if (ThisReadRP !=0 && opencount == 0) {
            Run((ULONG) ThisReadRP);/* dangling data                       */
            ThisReadRP=0L;
        }
        return (RPDONE);            /* return 'done' status to caller      */

    case RPREAD:                    /* 0x04                                */

        /*
        Try to read a character out of the queue.  If none are
        available, then Block until the interrupt routine can get a
        chance to add a character to the queue, and Run this thread.
        */

        ThisReadRP = rp;
        com_error_word=0;           /* start off with no errors            */
        ReadID = (ULONG) rp;
        if (Block(ReadID, -1L, 0, &err))
            if (err == 2)           /* interrupted                         */
                return(RPDONE|RPERR|ERROR_CHAR_CALL_INTERRUPTED);
        /*
        We have a single character (at least) from the queue.
        Move it to the user buffer, and while we are here, take
        any other characters available, up to the number of chars
        requested in the Read packet.  Return the final count to
        the caller.  This will always be one or more.
        */

        /* timed out */

        if (rx_queue.qcount == 0) { /* return 0 length                     */
           rp->s.ReadWrite.count=0;
           return (RPDONE|RPERR|ERROR_NOT_READY); /* tell kernel           */
           }

        i=0;
        do {
            if (MoveData(&inchar,
            (FARPOINTER) (rp->s.ReadWrite.buffer+i),
            1,
            MOVE_VIRTTOPHYS))
                return(RPDONE | RPERR | ERROR_GEN_FAILURE);
        }
        while (++i < rp->s.ReadWrite.count
            && !QueueRead(&rx_queue,&inchar));
        rp->s.ReadWrite.count = i;
        return(rp->RPstatus);

    case RPWRITE:                   /* 0x08                                */

        ThisWriteRP = rp;

        /* transfer characters from user buffer to write queue */

        addr=rp->s.ReadWrite.buffer;/* get address of callers buffer       */
        for (i = rp->s.ReadWrite.count; i; --i, ++addr)
        {
            if (MoveData((FARPOINTER)addr,
               &outchar,
               1,
               MOVE_PHYSTOVIRT))
                 return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            if (QueueWrite(&tx_queue,outchar))
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);
        }
        WriteID = (ULONG) rp;
        enable_write();

        if (Block(WriteID, -1L, 0, &err))
            if (err == 2)       /* interrupted                            */
                return(RPDONE|RPERR|ERROR_CHAR_CALL_INTERRUPTED);

        tickcount=MIN_TIMEOUT;    /* reset timeout value */
        return (rp->RPstatus);

    case RPINPUT_FLUSH:             /* 0x07                                */

        QueueFlush(&rx_queue);
        return (RPDONE);

    case RPOUTPUT_FLUSH:            /* 0x0b                                */

        QueueFlush(&tx_queue);
        return (RPDONE);

    case RPIOCTL:                   /* 0x10                                */

        /*
        The function code in an IOCtl packet has the high bit set
        for the serial board. We return all others (except 1) with the done
        bit set so we don't have to handle things like the 5-48
        code page IOCtl. Category 1 is the standard serial device category.
        This allows for the standard serial commands and programs like
        MODE to communicate with this driver.
        */

        if (!((rp->s.IOCtl.category == SAMPLE_CAT) ||
            (rp->s.IOCtl.category == 0x01))) /* just pass by other IOCtls */
            return (RPDONE);

        switch (rp->s.IOCtl.function)
        {
        case 0x41:                  /* set baud rate                       */

            /* set baud rate to 1200, 2400, 9600, 19200                    */

            /* verify caller owns the buffer area */

            if(VerifyAccess(
            SELECTOROF(rp->s.IOCtl.parameters), /* selector                */
            OFFSETOF(rp->s.IOCtl.parameters),   /* offset                  */
            2,                                  /* two bytes               */
            1) )                                /* read/write              */
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            /* lock the segment down temp */

            if(LockSeg(
            SELECTOROF(rp->s.IOCtl.parameters), /* selector                */
            0,                              /* lock for < 2 sec            */
            0,                              /* wait for seg lock           */
            (PLHANDLE) &lock_seg_han))      /* handle returned             */
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            /* get physical address of buffer */

            if (VirtToPhys(
            (FARPOINTER) rp->s.IOCtl.parameters,/* the virtual address     */
            (FARPOINTER) &appl_buffer))         /* physical address        */
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            /* move data to local driver buffer */

            if(MoveData(
            (FARPOINTER) appl_buffer, /* source                            */
            &baud_rate,             /* destination                         */
            2,                      /* 2 bytes                             */
            MOVE_PHYSTOVIRT))
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            if (UnPhysToVirt())
                return(RPDONE | RPERR | ERROR_GEN_FAILURE);

            /* unlock segment */

            if(UnLockSeg(lock_seg_han))
                return(RPDONE | RPERR | ERROR_GEN_FAILURE);

            switch (baud_rate)
            {
            case 1200:

                uart_regs.Bal=0xe0;
                uart_regs.Bah=0x01;
                break;

            case 2400:

                uart_regs.Bal=0xf0;
                uart_regs.Bah=0x00;
                break;

            case 9600:

                uart_regs.Bal=0x3c;
                uart_regs.Bah=0x00;
                break;

            case 19200:

                uart_regs.Bal=0x1e;
                uart_regs.Bah=0x00;
                break;

            case 38400:

                uart_regs.Bal=0x0f;
                uart_regs.Bah=0x00;
                break;

error:
                return (RPDONE | RPERR | ERROR_BAD_COMMAND);

            }
            init();                /* reconfigure uart                    */
            return (RPDONE);

        case 0x68:                  /* get number of chars in rec. queue   */

            /* verify caller owns the buffer area */

            if(VerifyAccess(
            SELECTOROF(rp->s.IOCtl.buffer), /* selector                    */
            OFFSETOF(rp->s.IOCtl.buffer),   /* offset                      */
            4,                              /* 4 bytes                     */
            1) )                            /* read/write                  */
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            /* lock the segment down temp */

            if(LockSeg(
            SELECTOROF(rp->s.IOCtl.buffer), /* selector                    */
            0,                              /* lock for < 2 sec            */
            0,                              /* wait for seg lock           */
            (PLHANDLE) &lock_seg_han))      /* handle returned             */
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            /* get physical address of buffer */

            if (VirtToPhys(
            (FARPOINTER) rp->s.IOCtl.buffer,/* the virtual address         */
            (FARPOINTER) &appl_buffer))     /* physical address            */
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            rqueue.cch=rx_queue.qcount;
            rqueue.cb=rx_queue.qsize;

            /* move data to user buffer */

            if(MoveData(
            &rx_queue,                      /* source                      */
            (FARPOINTER) appl_buffer,       /* destination                 */
            4,                              /* 4 bytes                     */
            MOVE_VIRTTOPHYS))
                return (RPDONE | RPERR | ERROR_GEN_FAILURE);

            if (UnPhysToVirt())
                return(RPDONE | RPERR | ERROR_GEN_FAILURE);

            /* unlock segment */

            if(UnLockSeg(lock_seg_han))
                return(RPDONE | RPERR | ERROR_GEN_FAILURE);

            return (RPDONE);

        default:
            return(RPDONE | RPERR | ERROR_GEN_FAILURE);
        }

        /* don't allow deinstall */

    case RPDEINSTALL:               /* 0x14                                */

        return(RPDONE | RPERR | ERROR_BAD_COMMAND);

        /* all other commands are ignored */

    default:
        return(RPDONE);

    }
}

void enable_write()                 /* enable write interrupts on uart     */
{
    int  port;
    int  reg_val;

    port=UART_PORT_ADDRESS;         /* get base port address               */
    reg_val=inp(port+2) & 0x60;     /* save current bank                   */  
    set_bank(0);                    /* switch to config bank               */
    outp((port+1),inp(port+1) | 0x12); /* turn on all write interrupts     */
    outp((port+2),reg_val);         /* restore bank info                   */

}
void disable_write()

/* turn off write interrupts on uart */

{
    int  port;
    int  reg_val;

    port=UART_PORT_ADDRESS;         /* get base port address               */
    reg_val=inp(port+2) & 0x60;     /* save current bank info              */
    set_bank(0);                    /* switch to config bank               */
    outp((port+1),inp(port+1) & 0xed); /* disable write interrupts         */
    outp((port+2),reg_val);         /* restore old bank                    */

}

void init ()

/* intializes software and configures 82050 */
{
    config_82050 ();                /* Configure 82050                     */
    set_bank(1);                    /* default operating bank = 1          */
}

void config_82050()

/*  Configure the 82050      */
{
    int  port;
    int inval;

    Disable();                      /* disable interrupts                  */
    port=UART_PORT_ADDRESS;

    /* set stick parity bit */

    set_bank(1);                    /* stick bit register                  */
    outp((port+7),0x10);            /* reset port                          */
    outp ((port+1), uart_regs.Txf); /* stick bit register                  */

    set_bank (2);                   /* general configuration bank          */
    outp ((port + 4), uart_regs.Imd);/* auto rupt ack                      */
    outp ((port + 7), uart_regs.Rmd);/* window,absolute start bit          */
    outp ((port + 5), uart_regs.Acr1);/* control-z                         */
    outp ((port + 3), uart_regs.Tmd);/* no 9 bit, no parity 2 sb's         */
    outp ((port + 1), uart_regs.Fmd);/* rx fifo threshold=1                */
    outp ((port + 6), uart_regs.Rie);/* enable receive interrupts          */

    set_bank (3);                   /* modem configuration bank            */

    outp ((port + 0), uart_regs.Clcf);/* clock source                      */
    set_dlab (3);               
    outp ((port + 0), uart_regs.Bbl);/* BRGB lsb                           */
    outp ((port + 1), uart_regs.Bbh);/* BRGB msb                           */
    reset_dlab (3);             
    outp ((port + 3), uart_regs.Bbcf);/* BRGB mode                         */
    outp ((port + 6), uart_regs.Tmie);/* timer b interrupt enable          */

    set_bank (0);                   /* general configuration               */

    outp ((port + 1), uart_regs.Ger);/* enable timer, interrupts           */
    outp ((port + 3), uart_regs.Lcr);/* 8 bit chars, no parity             */
    outp ((port + 7), uart_regs.Acr0);/* CR recognized                     */
    outp ((port + 4), uart_regs.Mcr_0);/* no DTR, RTS                      */
    set_dlab (0);               
    outp ((port + 0), uart_regs.Bal);/* BRGA lsb                           */
    outp ((port + 1), uart_regs.Bah);/* BRGA msb                           */
    reset_dlab (0);
    set_bank(1);

    Enable();                       /* turn interrupts back on              */
}

void set_dlab (bank)

/*  Set DLAB bit to allow access to divisior registers  */

int bank;
{
    int  inval;
    int  port;

    port=UART_PORT_ADDRESS;         /* get base port address               */
    set_bank (0);                   /* switch to gen gen config bank       */
    inval=inp(port +3);             /* set divisor latch access bit        */
    inval =inval | 0x80;            /* set dlab in LCR                     */
    outp ((port+3),inval);          /* do it                               */
    set_bank (bank);
}

getsrc()                            /* get source of interrupt             */
{
    int   v,src;
    int   port;

    port=UART_PORT_ADDRESS;         /* get base address of port            */
    v=inp(port+2);                  /* get data                            */
    src=v & 0x0e;                   /* mask bits                           */
    src=src/2;                      /* divide by 2                         */
    return(src);                    /* and pass it back                    */
}

set_bank(bank_num)

/* set bank of 82050 uart */

int  bank_num;
{
    int   reg_val;
    int   port;

    reg_val=bank_num*0x20;          /* select bank number to switch to     */
    port=UART_PORT_ADDRESS;         /* get real port address               */
    outp(port+gir_addr,reg_val);    /* output to correct port              */
}

void reset_dlab (bank)

/*  Reset DLAB bit of LCR   */

int bank;
{
    int  inval;
    int  port;

    port=UART_PORT_ADDRESS;         /* base port address                   */
    set_bank (0);                   /* switch to bank 0                    */
    inval=inp (port +3);            /* get divisor latch status            */
    inval = (inval & 0x7f);         /* dlab = 0 in LCR                     */
    outp ((port+3),inval);          /* reset it                            */
    set_bank (bank);
}

/* 82050 interrupt handler */

void interrupt_handler ()
{
    int  source;
    int  st_b;
    int  port;
    int  temp;
    int  rxlevel;


    port=UART_PORT_ADDRESS;
    set_bank(1);                     /* switch to bank 1                   */
    source = getsrc ();              /* get vector                         */
    switch (source)                  /* who dunnit?                        */
    {

    /* optional timer service routine */

    case timer :

        st_b=inp (port+3);    /* dec transmit counter     */
        if ( ThisReadRP == 0) /* nobody waiting      */
            break;
        ThisReadRP->RPstatus=(RPDONE | RPERR | ERROR_NOT_READY);
        Run (WriteID);/* run thread     */
        ThisWriteRP=0;
        break;

    case txm   :
    case txf   :

        /* spurious write interrupt */

        if ( ThisWriteRP == 0) {
            temp=inp(port+2);
            break;
        }

        /* keep transmitting until no data left */

        if  (!(QueueRead(&tx_queue,&outchar)))
        {
             outp((port), outchar);
             tickcount=MIN_TIMEOUT;
             break;
        }

        /* done writing, run blocked strategy thread */

        tickcount=MIN_TIMEOUT;
        disable_write();
        ThisWriteRP->RPstatus = (RPDONE);
        Run (WriteID);
        ThisWriteRP=0;
        break;

    case ccr   :

        /* control character, treat as normal character */

        inchar=inp(port+5);

    case rxf   :

        /* rx fifo service routine */

        if ( ThisReadRP == 0)
            inchar=inp (port); /* get character      */
        else
        {
        temp=inp(port+4);
        rxlevel=(temp & 0x70) / 0x10;

         /* empty out chip FIFO */

         while (rxlevel !=0) {

           inchar=inp (port); /* get character */
           rxlevel--;
           tickcount=MIN_TIMEOUT;

           /* write input data to circular queue */

           if(QueueWrite(&rx_queue,inchar))

             /* error, queue must be full */

             {
             ThisReadRP->RPstatus=(RPDONE|RPERR|ERROR_GEN_FAILURE);
             Run (ReadID);
             ThisReadRP=0;
             break;
             }
           com_error_word |= inp(port+5);

        } /* while rxlevel */
     } /* else */
  } /* switch (source) */
}

void timer_handler()
{
  if (ThisReadRP == 0)              /* nobody waiting to read              */
        return;

  tickcount--;
  if(tickcount == 0)  {
    ThisReadRP->RPstatus=(RPDONE | RPERR | ERROR_NOT_READY);
    Run (ReadID);
    ThisReadRP=0L;
    tickcount=MIN_TIMEOUT;
    }
}
