;----------------------------------------------------------------------;
;                                                                      ;
;    SPOOLER.ASM                                                       ;
;                                                                      ;
;       Copyright (c) 1991 Barry R. Nance                              ;
;       You may freely use this software without license fee           ;
;                                                                      ;
;       Distributed by Byte Magazine and the BIX network               ;
;                                                                      ;
;                                                                      ;
;USAGE: spooler <spoolsize> <priority> <formfeeds?> <outputfile>       ;
;                    |           |           |             |           ;
;                    |           |           |             |           ;
;    Spool File Size in K        |           |             |           ;
;                                |           |             |           ;
;          Background Print Priority         |             |           ;
;                                            |             |           ;
;            Insert FormFeeds between Printouts? (Y/N)     |           ;
;                                                          |           ;
;                         Spool to this file instead of to printer     ;
;                                                                      ;
;                                                                      ;
;       DEFAULTS:  spool file size........100K                         ;
;                  background priority.......5                         ;
;                  insert formfeeds..........N                         ;
;                  spool to file.............none.  Output to printer. ;
;                                                                      ;
;       <spoolsize>  can range from 5 to ???.                          ;
;       <priority>   can range from 1 to 20 (20 = fastest print).      ;
;       <formfeeds?> are turned on with a 'Y' (or 'y').                ;
;                    any other character turns FormFeeds off.          ;
;       <outputfile> can be any valid DOS filename.                    ;
;                                                                      ;
;       EXAMPLES:                                                      ;
;                                                                      ;
;       spooler 50 9 Y printer.fil                                     ;
;                                                                      ;
;       (this example sets up a 50K spool file on disk,                ;
;        sets the background print priority to 9,                      ;
;        tells SPOOLER to insert FormFeeds between printouts,          ;
;        and sends all output destined for the printer to PRINTER.FIL) ;
;                                                                      ;
;                                                                      ;
;       spooler 100 20 N                                               ;
;                                                                      ;
;       (this example sets up a 100K spool file,                       ;
;        sets the background priority to 20 (foreground will suffer!), ;
;        and tells SPOOLER not to insert FormFeeds.                    ;
;        Output will be spooled to the printer, not to a file.)        ;
;                                                                      ;
;                                                                      ;
; CAUTION: SPOOLER DOES NOT WORK WITH OUTPUT REDIRECTION (SEE THE      ;
;          EXAMPLE BELOW).  REDIRECTED OUTPUT MAY BE LOST.             ;
;                                                                      ;
;               dir >lpt1                                              ;
;                                                                      ;
;----------------------------------------------------------------------;


                .ALPHA
                .MODEL small
                Include DOS.Inc
                Include BIOS.Inc
StdIn           =       0000
StdOut          =       0001
StdErr          =       0002

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                .DATA
ControlBlock    Label   Byte
  SpoolSizePtr  dw      Offset  SpoolSize
  PriorityPtr   dw      Offset  Priority
  DoFFPtr       dw      Offset  DoFormFeeds
  PIDPtr        dw      Offset  OurPID
  OutputPtr     dw      Offset  OutputFile
  OPNamePtr     dw      Offset  FullName
  TimeSlicePtr  dw      Offset  TimeSliceTicks
  TotalCharsPtr dw      Offset  CharsInQueue
  DeInstallPtr  dw      Offset  DeInstall
  TogglePtr     dw      Offset  Toggle
  ResetSwPtr    dw      Offset  ResetSwitch
  EndOfCtl      Label   Byte

SpoolSize       dw      0
Priority        dw      0
PriorityHold    dw      0
IdleCounter     dw      0
DoFormFeeds     dw      0
ResetSwitch     dw      0
DeInstall       dw      0

OurPID          dw      0
OurPSP          dw      0
SpoolFile       dw      0
SpoolName       db      "spool.fil", 0
SpoolPathName   db      80 dup(0)

OutputFile      dw      0
OutputName      db      60 dup(0)
FullName        db      80 dup(0)
OutputLength    dw      0

KeepSize        dw      0

pname1          db      "LPT1"
                db      " "
pname2          db      "PRN"
                db      " "

PrintHandle     dw      -1
DupedHandle     dw      -1
last_char       db      0ch

OSMajor         db      0
OSMinor         db      0
ShareFlag       db      0
Low_Inten       db      0
Hi_Inten        db      0

i               db      0
j               db      0
k               dw      0
curloc          dw      0
curmode         dw      0

TickCounter     dd      0
TimeSliceTicks  dd      -1
NextSliceTicks  dd      0

WriteBufferFlag db      0
WeAreLast       db      0
Toggle          dw      0
BreakState      db      0
OldPID          dw      0

InDOS           Label   DWord
InDOS_Ofs       dw      0
InDOS_Seg       dw      0

InDOS2          Label   DWord
InDOS2_Ofs      dw      0
InDOS2_Seg      dw      0

OurDTA          Label   DWord
OurDTAOfs       dw      0
OurDTASeg       dw      0

OldDTA          Label   DWord
OldDTAOfs       dw      0
OldDTASeg       dw      0

OurMCB          Label   DWord
OurMCBOfs       dw      0
OurMCBSeg       dw      0

OurMCBSize      Label   DWord
OurMCBSizeOfs   dw      0
OurMCBSizeSeg   dw      0

Msg1            DB      13,10 
                DB      "Spooler is now loaded and running.", 13,10
                DB      "Copyright (c) 1991 Barry Nance", 13,10
                DB      "Use Spooler freely; there is no license fee.", 13,10
                DB      "Distributed by Byte Magazine and BIX", 13,10
LMsg1           =       $-Msg1

Msg2            DB      13,10
                DB      "Spooler was already loaded.", 13,10
LMsg2           =       $-Msg2

Msg3            DB      13,10
                DB      "Early versions of DOS not supported.", 13,10
LMsg3           =       $-Msg3

Msg4            DB      13,10
                DB      "ERROR.  Could not create spool file.", 13,10
LMsg4           =       $-Msg4

Msg5            DB      13,10
                DB      "Creating spool file.  Please wait...", 13,10
LMsg5           =       $-Msg5

Msg6            DB      13,10
                DB      "Spool file successfully created.", 13,10
LMsg6           =       $-Msg6

Msg7            DB      13,10
                DB      "ERROR.  Could not create output file.", 13,10
LMsg7           =       $-Msg7

Msg8            DB      13,10
   DB   "USAGE: Spooler <buffsize> <priority> <formfeeds?> [diskfile]", 13,10
                DB      "                 DEFAULT    RANGE", 13,10
                DB      " buffer size       100K     5-999",  13,10
                DB      " priority            5      1-20 ",  13,10
                DB      " insert formfeeds    N       Y/N ",  13,10
                DB      " diskfile name    <none>", 13,10
                DB      13,10
LMsg8           =       $-Msg8

ReadChars       dw      0
WriteChars      dw      0
BuffChars       dw      0
OutputChars     dw      0

SpoolPosition   dw      0
OutputPosition  dw      0
CharsInQueue    dd      0
WritePosition   dd      0
ReadPosition    dd      0
FileSize        dd      0

WriteBuffer     DB      512 dup(0h)
ReadBuffer      DB      512 dup(0h)

String1         DB      128 dup(0h)

HiVal           DW      0
LoVal           DW      0
Sign            DW      0

IntTable        DD           1
                DD          10
                DD         100
                DD        1000
                DD       10000
                DD      100000
                DD     1000000
                DD    10000000
                DD   100000000

TempString      db   34 dup(0)

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
                .STACK  100h
; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

                .CODE
func_id         db      0
app_dx          dw      0
app_ds          dw      0

flags           dw      0
save_ah         db      0
save_ss         dw      0
save_sp         dw      0
current_ss      dw      0
current_sp      dw      0

In_Int08        db      0
In_Int10        db      0
In_Int13        db      0
In_Int16        db      0
In_Int21        db      0
In_Int28        db      0
InBackground    db      0
DataToProcess   db      0
PrtScFlag       db      0
DontSpool       db      0

StdPrn          dw      0004

OldInt05        Label   DWord
OldInt05Ofs     DW      0
OldInt05Seg     DW      0

OldInt08        Label   DWord
OldInt08Ofs     DW      0
OldInt08Seg     DW      0

OldInt10        Label   DWord
OldInt10Ofs     DW      0
OldInt10Seg     DW      0

OldInt13        Label   DWord
OldInt13Ofs     DW      0
OldInt13Seg     DW      0

OldInt16        Label   DWord
OldInt16Ofs     DW      0
OldInt16Seg     DW      0

OldInt17        Label   DWord
OldInt17Ofs     DW      0
OldInt17Seg     DW      0

OldInt20        Label   DWord
OldInt20Ofs     DW      0
OldInt20Seg     DW      0

OldInt21        Label   DWord
OldInt21Ofs     DW      0
OldInt21Seg     DW      0

OldInt23        Label   DWord
OldInt23Ofs     DW      0
OldInt23Seg     DW      0

OldInt24        Label   DWord
OldInt24Ofs     DW      0
OldInt24Seg     DW      0

OldInt28        Label   DWord
OldInt28Ofs     DW      0
OldInt28Seg     DW      0

Seg40           dw      0040h

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_05:         STI
                Cmp     DontSpool, 1
                JNE     Int05_10
                CLI
                Jmp     DWord Ptr OldInt05

Int05_10:       Push    ES

                Push    Seg40
                Pop     ES

                Cmp     Byte Ptr ES:[0049h], 3
                JE      Int05_30
                Cmp     Byte Ptr ES:[0049h], 7
                JE      Int05_30

                Pop     ES
                IRet

Int05_30:       Pop     ES
                Mov     DataToProcess, 1
                Mov     PrtScFlag, 1
                IRet

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_08:         Mov     In_Int08, 1
                Pushf
                Call    OldInt08
                STI

                push    ax
                push    bx
                push    cx
                push    dx
                push    es
                push    ds
                push    si
                push    di
                push    bp

                mov     bp, @DATA
                mov     ds, bp

                Add     word ptr TickCounter, 1
                Adc     word ptr TickCounter+2, 0

                Cmp     ResetSwitch, 1
                JNE     CheckFlags

                Mov     ResetSwitch, 0
                Mov     ReadChars, 0
                Mov     WriteChars, 0
                Mov     BuffChars, 0
                Mov     SpoolPosition, 0
                Mov     OutputPosition, 0
                Mov     word ptr CharsInQueue, 0
                Mov     word ptr CharsInQueue+2, 0
                Mov     word ptr ReadPosition, 0
                Mov     word ptr ReadPosition+2, 0
                Mov     word ptr WritePosition, 0
                Mov     word ptr WritePosition+2, 0

CheckFlags:     Cmp     DataToProcess, 1
                JE      I08_10
                Cmp     DeInstall, 1
                JE      I08_10

                Cmp     Toggle, 1
                JE      TurnOff
                Jmp     TurnOn

TurnOff:        Mov     DontSpool, 1
                Jmp     Int08Exit

TurnOn:         Mov     DontSpool, 0
                Jmp     Int08Exit

I08_10:         Cmp     In_Int10, 1
                JE      Int08Exit
                Cmp     In_Int13, 1
                JE      Int08Exit
                Cmp     In_Int16, 1
                JE      Int08Exit
                Cmp     In_Int28, 1
                JE      Int08Exit
                Cmp     In_Int21, 1
                JE      Int08Exit

                Cmp     InBackground, 1
                JE      Int08Exit

                LES     BX, InDOS
                Cmp     Byte Ptr ES:[BX], 0
                JNE     Int08Exit
                LES     BX, InDOS2
                Cmp     Byte Ptr ES:[BX], 0
                JNE     Int08Exit

                Mov     AL, 0bh
                Out     20h, AL
                In      AL, 20h
                Cmp     AL, 0
                JNE     Int08Exit

                Mov     InBackground, 1
                Call    DoPopUp
                Mov     InBackground, 0

Int08Exit:      pop     bp
                pop     di
                pop     si
                pop     ds
                pop     es
                pop     dx
                pop     cx
                pop     bx
                pop     ax
                mov     In_Int08, 0
                IRet

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_10:         Mov     In_Int10, 1
                STI
                Pushf
                CLI
                Call    OldInt10
                Mov     In_Int10, 0
                Retf    2

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_13:         Mov     In_Int13, 1
                STI
                Pushf
                CLI
                Call    OldInt13
                Mov     In_Int13, 0
                Retf    2

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_16:         Mov     In_Int16, 1
                STI

                Cmp     AX, "BN"
                JNE     I16_10

                Mov     AX, @DATA
                Mov     ES, AX
                Mov     BX, Offset ControlBlock
                Mov     AX, "bn"
                Mov     In_Int16, 0
                IRet

I16_10:         push    ax
                push    bx
                push    cx
                push    dx
                push    es
                push    ds
                push    si
                push    di
                push    bp

                mov     bp, @DATA
                mov     ds, bp

                Cmp     InBackground, 1
                JE      DoOld16

                Cmp     AH, 0
                JNE     DoOld16

WaitForKey:     Mov     AH, 1
                PushF
                CLI
                Call    OldInt16
                JZ      Popup16
                Jmp     DoOld16

Popup16:        Inc     IdleCounter
                Cmp     DataToProcess, 1
                JE      I16_30
                Cmp     DeInstall, 1
                JNE     WaitForKey

I16_30:         LES     BX, InDOS
                Cmp     Byte Ptr ES:[BX], 0
                JNE     WaitForKey
                LES     BX, InDOS2
                Cmp     Byte Ptr ES:[BX], 0
                JNE     WaitForKey

                Mov     InBackground, 1
                Call    DoPopUp
                Mov     InBackground, 0
                Jmp     WaitForKey

DoOld16:        pop     bp
                pop     di
                pop     si
                pop     ds
                pop     es
                pop     dx
                pop     cx
                pop     bx
                pop     ax

                PushF
                CLI
                Call    OldInt16
                Mov     In_Int16, 0

                RetF    2

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_17:         STI
                push    ax
                push    bx
                push    cx
                push    dx
                push    es
                push    ds
                push    si
                push    di
                push    bp

                mov     bp, @DATA
                mov     ds, bp

                Cmp     DontSpool, 1
                JE      DoOld17
                Cmp     DX, 0
                JNE     DoOld17

                Mov     save_ah, 90h
                Cmp     AH, 0
                JNE     Int17Exit

                LES     BX, InDOS
                Cmp     Byte Ptr ES:[BX], 0
                JNE     DoOld17
                LES     BX, InDOS2
                Cmp     Byte Ptr ES:[BX], 0
                JNE     DoOld17

                Call    Queue
                Jmp     Int17Exit

DoOld17:        PushF
                CLI
                Call    OldInt17
                Mov     save_ah, ah

Int17Exit:      pop     bp
                pop     di
                pop     si
                pop     ds
                pop     es
                pop     dx
                pop     cx
                pop     bx
                pop     ax

                mov     ah, save_ah
                IRet

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_20:         STI
                Mov     AX, @DATA
                Mov     DS, AX

                mov     StdPrn, 0004

I20_10:         cmp     PrintHandle, -1
                je      I20_30
                cmp     last_char, 0ch
                je      I20_30

                mov     al, 0ch
                call    Queue

I20_30:         mov     PrintHandle, -1
                Mov     AX, 4c00h
                PushF
                CLI
                Call    OldInt21

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_21:         cmp     ah, 4bh
                jne     not_exec
                cmp     al, 03
                je      GotoDOS

                mov     StdPrn, 0004

GotoDOS:        jmp     dword ptr OldInt21

not_exec:       cmp     ah, 25h
                je      GotoDOS
                cmp     ah, 35h
                je      GotoDOS
                cmp     ah, 1ah
                je      GotoDOS
                cmp     ah, 2fh
                je      GotoDOS

                Cmp     DontSpool, 1
                JE      GotoDOS
                Cmp     InBackground, 1
                JE      GotoDOS

                mov     in_int21, 1
                sti

                clc
                pushf
                pop     flags

                mov     app_ds, ds
                mov     app_dx, dx
                mov     func_id, ah

                push    ax
                push    bx
                push    cx
                push    dx
                push    es
                push    ds
                push    si
                push    di
                push    bp

                mov     bp,@DATA
                mov     ds,bp
                mov     bp, sp
                cld

@20:            cmp     ah, 22h
                je      @25
                cmp     ah, 15h
                je      @25

                jmp     @66

@25:            mov     es, app_ds
                mov     di, app_dx
                cmp     byte ptr es:[di], 0ffh
                jne     @28

                add     di, 7

@28:            inc     di
                mov     si, offset pname1
                mov     cx, 5
                push    di
        repe    cmpsb
                pop     di
                je      @30

                mov     si, offset pname2
                mov     cx, 4
                push    di
        repe    cmpsb
                pop     di
                je      @30

                jmp     @47

@30:            mov     cx, word ptr es:[di+13]
                xor     si,si

                mov     ah, 2fh
                pushf
                cli
                call    OldInt21

                jmp     short @36

@33:            mov     al, byte ptr es:[bx]

                push    bx
                push    cx
                push    si
                push    es
                call    Queue
                pop     es
                pop     si
                pop     cx
                pop     bx

                inc     bx
                inc     si

@36:            cmp     si, cx
                jb      @33

@38:            mov     byte ptr [bp+17], 00
                jmp     Int21Exit




@66:            cmp     ah, 05h
                jne     @44

                mov     al, dl
                call    Queue
                jmp     Int21Exit

@44:            cmp     ah, 40h
                jne     @47

                cmp     bx, StdPrn
                je      @48

chk_prthandle:  cmp     bx, DupedHandle
                je      @48
                cmp     bx, PrintHandle
                jne     @47

@48:            xor     si,si
                mov     es, app_ds
                mov     bx, app_dx
                jmp     short @52

@51:            mov     al, byte ptr es:[bx]

                push    bx
                push    cx
                push    si
                push    es
                call    Queue
                pop     es
                pop     si
                pop     cx
                pop     bx

                inc     bx
                inc     si

@52:            cmp     si, cx
                jb      @51

@45:            mov     word ptr [bp+16], cx
                jmp     Int21Exit

@47:            cmp     ah, 00h
                je      @47_A
                cmp     ah, 31h
                je      @54
                cmp     ah, 4ch
                jne     @53

@47_A:          mov     func_id, 4ch

@54:            cmp     PrintHandle, -1
                je      @55
                cmp     last_char, 0ch
                je      @55

                mov     al, 0ch
                call    Queue

@55:            mov     PrintHandle, -1
                mov     DupedHandle, -1
                mov     StdPrn, 0004

                pop     bp
                pop     di
                pop     si
                pop     ds
                pop     es
                pop     dx
                pop     cx
                pop     bx
                pop     ax

                mov     ah, func_id
                mov     in_int21, 0
                cli     
                jmp     dword ptr OldInt21


@53:            pop     bp
                pop     di
                pop     si
                pop     ds
                pop     es
                pop     dx
                pop     cx
                pop     bx
                pop     ax

                pushf   
                cli     
                call    OldInt21

                pushf
                pop     flags
        
                push    ax
                push    bx
                push    cx
                push    dx
                push    es
                push    ds
                push    si
                push    di
                push    bp

                mov     bp, @DATA
                mov     ds, bp
                mov     bp, sp
                cld

                cmp     func_id, 31h
                ja      @60

                and     flags, 0fffeh
                jmp     Int21Exit

@60:            test    flags, 1
                jz      carry_func
                jmp     Int21Exit

carry_func:     cmp     func_id, 3eh
                jne     @61

                cmp     bx, DupedHandle
                jne     chk_prn
                mov     DupedHandle, -1

chk_prn:        cmp     bx, PrintHandle
                je      lpt1_close

                cmp     bx, StdPrn
                je      PrnClose

                jmp     @40

PrnClose:       cmp     last_char, 0ch
                je      no_ff

                mov     al, 0ch
                call    Queue

no_ff:          mov     StdPrn, -1
                jmp     @40

lpt1_close:     cmp     last_char, 0ch
                je      no_need

                mov     al, 0ch
                call    Queue

no_need:        mov     PrintHandle, -1
                jmp     @40


@61:            cmp     func_id, 46h
                jne     @62

                cmp     bx, StdPrn
                je      forced_dupe

chk_force:      cmp     bx, PrintHandle
                je      forced_dupe
                jmp     @40

forced_dupe:    mov     DupedHandle, CX
                jmp     @40

@62:            cmp     func_id, 45h
                jne     open_or_close

                cmp     bx, StdPrn
                je      duped_print

chk_dupe:       cmp     bx, PrintHandle
                jne     open_or_close

duped_print:    mov     DupedHandle, AX
                jmp     @40

open_or_close:  cmp     func_id, 3dh
                je      @63
                cmp     func_id, 3ch
                je      @63

                jmp     @40

@63:            push    ax
                mov     es, app_ds
                mov     di, app_dx
                mov     al, 0
                mov     cx, 128
        repne   scasb
                pop     ax

back_thru_name: dec     di
                cmp     byte ptr es:[di], "\"
                je      found_start
                cmp     byte ptr es:[di], "/"
                je      found_start
                cmp     byte ptr es:[di], ":"
                je      found_start

                cmp     byte ptr es:[di], "_"
                jbe     first_pos_test
                and     byte ptr es:[di], "_"

first_pos_test: cmp     di, app_dx
                jne     back_thru_name

                dec     di

found_start:    inc     di

                mov     cx, 4
                mov     si, offset pname1
                push    di
         repe   cmpsb
                pop     di
                je      @64

                mov     cx, 3
                mov     si, offset pname2
                push    di
         repe   cmpsb
                pop     di
                je      @64

                jmp     short @40

@64:            mov     PrintHandle, ax

@40:            Cmp     DataToProcess, 1
                JE      I21_10
                Cmp     DeInstall, 1
                JNE     Int21Exit

I21_10:         LES     BX, InDOS
                Cmp     Byte Ptr ES:[BX], 0
                JNE     Int21Exit
                LES     BX, InDOS2
                Cmp     Byte Ptr ES:[BX], 0
                JNE     Int21Exit

                Mov     InBackground, 1
                Call    DoPopUp
                Mov     InBackground, 0

Int21Exit:      pop     bp
                pop     di
                pop     si
                pop     ds
                pop     es
                pop     dx
                pop     cx
                pop     bx
                pop     ax

                mov     in_int21,0
                push    flags
                popf
                retf    2

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_28:         Mov     In_Int28, 1
                Pushf
                Call    OldInt28
                STI

                push    ax
                push    bx
                push    cx
                push    dx
                push    es
                push    ds
                push    si
                push    di
                push    bp

                mov     bp, @DATA
                mov     ds, bp

                Cmp     DataToProcess, 1
                JE      I28_10
                Cmp     DeInstall, 1
                JNE     Int28Exit

I28_10:         Cmp     InBackground, 1
                JE      Int28Exit

                Inc     IdleCounter

                LES     BX, InDOS
                Cmp     Byte Ptr ES:[BX], 1
                JA      Int28Exit
                LES     BX, InDOS2
                Cmp     Byte Ptr ES:[BX], 0
                JNE     Int28Exit

                Mov     InBackground, 1
                Call    DoPopUp
                Mov     InBackground, 0

Int28Exit:      pop     bp
                pop     di
                pop     si
                pop     ds
                pop     es
                pop     dx
                pop     cx
                pop     bx
                pop     ax
                mov     In_Int28, 0
                IRet

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_23:         IRet

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Int_24:         STI
                Pop     AX
                Pop     AX
                Pop     AX

                Pop     AX
                Pop     BX
                Pop     CX
                Pop     DX
                Pop     SI
                Pop     DI
                Pop     BP
                Pop     DS
                Pop     ES

                Mov     AX, 83
                STC
                RetF    2

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

DoPopup         Proc    Near
                CLI
                Mov     current_ss, SS
                Mov     current_sp, SP
                Mov     SS, save_ss
                Mov     SP, save_sp
                STI
                CLD

                Mov     AX, OurPSP
                LES     BX, OurMCBSize
                Add     AX, ES:[BX]
                Mov     ES, AX
                Mov     AX, ES:[0001]

                Cmp     AX, 0000
                JE      SetLast
                Cmp     AX, 0ffffh
                JE      SetLast
                Cmp     AX, OurPSP
                JB      SetLast
                Jmp     Short SetNotLast

SetLast:        Mov     WeAreLast, 1
                Jmp     Short CheckDeInstall

SetNotLast:     Mov     WeAreLast, 0

CheckDeInstall: Cmp     DeInstall, 1
                JE      StartDeInstall

                Jmp     ProcessPopup

StartDeInstall: Mov     AX, 5100h
                PushF
                CLI
                Call    OldInt21
                Mov     OldPID, BX

                Mov     AX, 5000h
                Mov     BX, OurPID
                PushF
                CLI
                Call    OldInt21

                Cmp     DontSpool, 1
                JE      CheckIfLast

                Mov     DontSpool, 1

CloseSpool:     Mov     BX, SpoolFile
                Mov     AH, 3eh
                PushF
                CLI
                Call    OldInt21

EraseSpoolFile: Mov     DX, Offset SpoolPathName
                Mov     AH, 41h
                PushF
                CLI
                Call    OldInt21



CheckIfLast:    Cmp     WeAreLast, 1
                JE      RemoveISRs

                Mov     AX, 5000h
                Mov     BX, OldPID
                PushF
                CLI
                Call    OldInt21

                Jmp     DoPopupExit



RemoveISRs:     Push    DS
                Mov     DX, Word Ptr OldInt08
                Mov     DS, Word Ptr OldInt08+2
                Mov     AX, 2508h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt28
                Mov     DS, Word Ptr OldInt28+2
                Mov     AX, 2528h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt21
                Mov     DS, Word Ptr OldInt21+2
                Mov     AX, 2521h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt20
                Mov     DS, Word Ptr OldInt20+2
                Mov     AX, 2520h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt17
                Mov     DS, Word Ptr OldInt17+2
                Mov     AX, 2517h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt16
                Mov     DS, Word Ptr OldInt16+2
                Mov     AX, 2516h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt13
                Mov     DS, Word Ptr OldInt13+2
                Mov     AX, 2513h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt10
                Mov     DS, Word Ptr OldInt10+2
                Mov     AX, 2510h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                Mov     DX, Word Ptr OldInt05
                Mov     DS, Word Ptr OldInt05+2
                Mov     AX, 2505h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Mov     AX, CS
                Mov     ES, AX
                Mov     AH, 49h
                Int     21h

                Mov     ES, OurPSP
                Mov     ES, ES:[2ch]
                Mov     AH, 49h
                Int     21h

                Mov     AX, 4c00h
                Int     21h

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

ProcessPopup:   Mov     AX, 3300h
                PushF
                CLI
                Call    OldInt21
                Mov     BreakState, DL

                Mov     AX, 3523h
                PushF
                CLI
                Call    OldInt21
                Mov     OldInt23Seg, ES
                Mov     OldInt23Ofs, BX

                Push    DS
                Mov     DX, Offset Int_23
                Push    CS
                Pop     DS
                Mov     AX, 2523h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Mov     AX, 3524h
                PushF
                CLI
                Call    OldInt21
                Mov     OldInt24Seg, ES
                Mov     OldInt24Ofs, BX

                Push    DS
                Mov     DX, Offset Int_24
                Push    CS
                Pop     DS
                Mov     AX, 2524h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Mov     AH, 2fh
                PushF
                CLI
                Call    OldInt21
                Mov     OldDTASeg, ES
                Mov     OldDTAOfs, BX

                Push    DS
                Mov     AH, 1ah
                LDS     DX, OurDTA
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Mov     AX, 5100h
                PushF
                CLI
                Call    OldInt21
                Mov     OldPID, BX

                Mov     AX, 5000h
                Mov     BX, OurPID
                PushF
                CLI
                Call    OldInt21


                Call    ActualPopup


                Mov     AX, 5000h
                Mov     BX, OldPID
                PushF
                CLI
                Call    OldInt21

                Push    DS
                Mov     AH, 1ah
                LDS     DX, OldDTA
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                LDS     DX, OldInt24
                Mov     AX, 2524h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Push    DS
                LDS     DX, OldInt23
                Mov     AX, 2523h
                PushF
                CLI
                Call    OldInt21
                Pop     DS

                Mov     DL, BreakState
                Mov     AX, 3301h
                PushF
                CLI
                Call    OldInt21

DoPopupExit:
                CLI
                Mov     SS, current_ss
                Mov     SP, current_sp
                STI

                Ret
DoPopup         EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

ErrorNoise      Proc    Near
                Call    Beep
                Mov     BX, 2
                Call    Pause
                Call    Beep
                Mov     BX, 1
                Call    Pause
                Call    Beep
                Mov     BX, 2
                Call    Pause
                Call    Beep
                Ret
ErrorNoise      EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Queue           Proc    Near
                Push    AX

LetPrinterCatchUp:
                Mov     DX, word ptr WritePosition+2
                Mov     AX, word ptr WritePosition
                Mov     CX, word ptr ReadPosition+2
                Mov     BX, word ptr ReadPosition

CalcGap:        Cmp     DX, CX
                JA      Wrt_Minus_Rd
                JB      Sz_Minus_Rd_Plus_Wrt
                Cmp     AX, BX
                JAE     Wrt_Minus_Rd
                Jmp     Sz_Minus_Rd_Plus_Wrt

Wrt_Minus_Rd:   Sub     AX, BX
                Sbb     DX, CX
                Jmp     CheckForOverrun

Sz_Minus_Rd_Plus_Wrt:
                Mov     DI, word ptr FileSize+2
                Mov     SI, word ptr FileSize
                Sub     SI, BX
                Sbb     DI, CX
                Add     AX, SI
                Adc     DX, DI

CheckForOverrun:
                Mov     DI, word ptr FileSize+2
                Mov     SI, word ptr FileSize
                Sub     SI, 512
                Sbb     DI, 0
                Cmp     DX, DI
                JA      HaveToWait
                JB      PutCharInQueue
                Cmp     AX, SI
                JB      PutCharInQueue

HaveToWait:     Cmp     OutputFile, 1
                JA      LetPrinterCatchUp

                Mov     AH, 2
                Mov     DX, 0
                PushF
                CLI
                Call    OldInt17
                And     AH, 39h
                Xor     AH, 10h
                Cmp     AH, 0
                JE      LetPrinterCatchUp

                Call    ErrorNoise
                Jmp     LetPrinterCatchUp

PutCharInQueue: Pop     AX
                Mov     in_int21, 0
                Inc     BuffChars

                Mov     DataToProcess, 0

                Mov     SI, SpoolPosition
                Mov     WriteBuffer[SI], AL
                Inc     SI
                Mov     SpoolPosition, SI

                Mov     DataToProcess, 1
                Mov     WriteBufferFlag, 1

                Cmp     DoFormFeeds, 1
                JNE     Q_20
                Mov     last_char, AL

Q_20:           Cmp     BuffChars, 512
                JE      Q_SetWait
                Ret

Q_SetWait:      CLI
                Mov     DX, word ptr TickCounter
                Mov     CX, word ptr TickCounter+2
                STI
                Add     DX, 91
                Adc     CX, 0

Q_Wait:         STI
                Cmp     WriteBufferFlag, 1
                JNE     Q_99

                CLI
                Cmp     CX, word ptr TickCounter+2
                JA      Q_Wait
                JB      Q_Timeout
                Cmp     DX, word ptr TickCounter
                JA      Q_Wait

Q_Timeout:      STI
                Call    ErrorNoise
                Jmp     Q_SetWait

Q_90:           Mov     DataToProcess, 0
                Mov     WriteBufferFlag, 0

Q_99:           Ret
Queue           EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

ActualPopup     Proc    Near
                Mov     DataToProcess, 0
                Mov     ReadChars, 0

AnythingToWrite:
                Cmp     WriteBufferFlag, 1
                JNE     IsItTimeToPrint

                Mov     AX, BuffChars
                Add     word ptr CharsInQueue, AX
                Adc     word ptr CharsInQueue+2, 0
                Mov     WriteChars, AX
                Mov     BuffChars, 0
                Mov     SpoolPosition, 0
                Call    WriteToSpoolFile
                Mov     WriteBufferFlag, 0

IsItTimeToPrint:
                Mov     PriorityHold, 1
                Cmp     IdleCounter, 8
                JAE     AnythingToPrint

                Mov     BX, Priority
                Mov     PriorityHold, BX

                CLI
                Mov     AX, word ptr TickCounter+2
                Mov     BX, word ptr TickCounter
                STI
                Cmp     AX, word ptr NextSliceTicks+2
                JA      AnythingToPrint
                JB      NotThisTime
                Cmp     BX, word ptr NextSliceTicks
                JA      AnythingToPrint
 
NotThisTime:    Mov     DataToProcess, 1
                Jmp     AfterActualPrint

AnythingToPrint:
                Mov     IdleCounter, 0

                Cmp     word ptr CharsInQueue+2, 0
                JNE     CheckPrinter
                Cmp     word ptr CharsInQueue, 0
                JNE     CheckPrinter
                Jmp     AfterActualPrint

CheckPrinter:   Mov     DataToProcess, 1
                Cmp     OutputFile, 1
                JA      PrinterIsOnline

AP_10:          Mov     AH, 2
                Mov     DX, 0
                PushF
                CLI
                Call    OldInt17
                And     AH, 39h
                Xor     AH, 10h
                Cmp     AH, 0
                JNE     AP_20
                Jmp     PrinterIsOnline

AP_20:          Call    ErrorNoise
                Jmp     AfterActualPrint

PrinterIsOnline:
                CLI
                Mov     AX, word ptr TickCounter
                Mov     BX, word ptr TickCounter+2
                STI
                Mov     word ptr TimeSliceTicks, AX
                Mov     word ptr TimeSliceTicks+2, BX

AP_30:          Mov     BX, PriorityHold

DoubleTime:     Cmp     WeAreLast, 1
                JNE     AP_35
                Shl     BX, 1
                Cmp     BX, 10
                JBE     AP_35
                Mov     BX, 10

AP_35:          Add     word ptr TimeSliceTicks, BX
                Adc     word ptr TimeSliceTicks+2, 0

AP_TimeChk:     CLI
                Mov     AX, word ptr TickCounter+2
                Mov     BX, word ptr TickCounter
                STI
                Cmp     AX, word ptr TimeSliceTicks+2
                JB      ActualPrint
                JA      SetNextSlice
                Cmp     BX, word ptr TimeSliceTicks
                JB      ActualPrint
 
SetNextSlice:   CLI
                Mov     AX, word ptr TickCounter+2
                Mov     BX, word ptr TickCounter
                STI
                Mov     CX, 18
                Sub     CX, Priority
                Add     BX, CX
                Adc     AX, 0
                Mov     word ptr NextSliceTicks+2, AX
                Mov     word ptr NextSliceTicks, BX
                Jmp     AfterActualPrint

ActualPrint:    Cmp     word ptr CharsInQueue+2, 0
                JNE     ChkNeedBuff
                Cmp     word ptr CharsInQueue, 0
                JNE     ChkNeedBuff

                Mov     DataToProcess, 0
                Jmp     AfterActualPrint

ChkNeedBuff:    Cmp     ReadChars, 0
                JNE     PrintSomething

                Call    ReadFromSpool

PrintSomething: Mov     SI, OutputPosition

                Cmp     OutputFile, 1
                JB      ToPrinter

ToOutputFile:   Mov     AX, ReadChars
                Mov     OutputChars, AX

OutToDisk:      Lea     DX, ReadBuffer[SI]
                Mov     CX, OutputChars
                Mov     BX, OutputFile
                Mov     AX, 4000h
                PushF
                CLI
                Call    OldInt21
                JC      OutputError
                Cmp     AX, CX
                JNE     OutputError

                Mov     BX, 0
                Jmp     DoneWithChars

OutputError:    Call    ErrorNoise
                Mov     BX, 10
                Call    Pause
                Call    ErrorNoise
                Mov     BX, 10
                Call    Pause
                Call    ErrorNoise
                Mov     BX, 0
                Jmp     DoneWithChars

ToPrinter:      Mov     OutputChars, 1
                Mov     AL, ReadBuffer[SI]
                Mov     AH, 0
                Push    AX
                Mov     DX, 0
                PushF
                CLI
                Call    OldInt17
                Pop     BX

WasItOkay:      And     AH, 39h
                Xor     AH, 10h
                Cmp     AH, 0
                JE      DoneWithChars

                Call    ErrorNoise

DoneWithChars:  Mov     AX, OutputChars
                Sub     word ptr CharsInQueue, AX
                Sbb     word ptr CharsInQueue+2, 0

                Add     OutputPosition, AX

                Mov     AX, ReadChars
                Cmp     AX, OutputPosition
                JA      NextPageTest

                Mov     ReadChars, 0

NextPageTest:   Cmp     BL, 0ch
                JE      AfterActualPrint

GoCheckForMore: Jmp     AP_TimeChk


AfterActualPrint:
                Cmp     PrtScFlag, 1
                JNE     CheckBuffer

                Call    DoPrtSc

CheckBuffer:    Cmp     OutputFile, 1
                JB      ActualPopupExit

FlushOutput:    Mov     AH, 45h
                Mov     BX, OutputFile
                PushF
                CLI
                Call    OldInt21
                JC      ActualPopupExit

                Mov     BX, AX
                Mov     AH, 3Eh
                PushF
                CLI
                Call    OldInt21

ActualPopupExit:
                Ret
ActualPopup     EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

DoPrtSc         Proc    Near
                Mov     PrtScFlag, 0
                Mov     AH, 0fh
                Int     10h
                Mov     BP, BX

                Mov     AH, 3
                Int     10h
                Mov     curloc, DX
                Mov     curmode, CX

                Mov     i, 0
itest:          Cmp     i, 25
                JNE     resetj
                Jmp     EndOfPrtSc

resetj:         Mov     j, 0

jtest:          Cmp     j, 80
                JE      EndOfRow

                Mov     DH, i
                Mov     DL, j
                Mov     AH, 2
                Mov     BX, BP
                Int     10h
                Mov     AH, 8
                Int     10h
                
                Call    AL_To_WriteBuff

EndOfChar:      Inc     j
                Jmp     jtest

EndOfRow:       Mov     AL, 0dh
                Call    AL_To_WriteBuff

DoLFChar:       Mov     AL, 0ah
                Call    AL_To_WriteBuff

NextPrtScRow:   Inc     i
                Jmp     itest

EndOfPrtSc:     Mov     DX, curloc
                Mov     AH, 2
                Mov     BX, BP
                Int     10h
                Mov     CX, curmode
                Mov     AH, 1
                Int     10h

                Mov     DataToProcess, 1
                Ret
DoPrtSc         EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

AL_To_WriteBuff Proc    Near
                Inc     WriteChars

                Add     word ptr CharsInQueue, 1
                Adc     word ptr CharsInQueue+2, 0

                Mov     SI, SpoolPosition
                Mov     WriteBuffer[SI], AL
                Call    WriteToSpoolFile
                Mov     SpoolPosition, 0

AL_Exit:        Ret
AL_To_WriteBuff EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

ReadFromSpool   Proc    Near
SeekForRead:    Mov     CX, word ptr ReadPosition+2
                Mov     DX, word ptr ReadPosition
                Mov     AX, 4200h
                Mov     BX, SpoolFile
                PushF
                CLI
                Call    OldInt21

ReadHowMany:    Mov     ReadChars, 512
                Cmp     word ptr CharsInQueue+2, 0
                JA      ReadIt
                Cmp     word ptr CharsInQueue, 512
                JA      ReadIt

                Mov     AX, word ptr CharsInQueue
                Mov     ReadChars, AX

ReadIt:         Mov     AX, 3F00h
                Mov     BX, SpoolFile
                Mov     CX, ReadChars
                Mov     DX, Offset ReadBuffer
                PushF
                CLI
                Call    OldInt21
                JC      ReadError

SetForNextRead: Mov     ReadChars, AX
                Mov     CX, AX
                Mov     AX, word ptr ReadPosition
                Mov     DX, word ptr ReadPosition+2
                Call    IncPosition
                Mov     word ptr ReadPosition, AX
                Mov     word ptr ReadPosition+2, DX

ReadExit:       Mov     OutputPosition, 0
                Ret

ReadError:      Call    ErrorNoise
                Mov     BX, 10
                Call    Pause
                Call    ErrorNoise
                Mov     BX, 10
                Call    Pause
                Call    ErrorNoise
                Jmp     ReadExit

ReadFromSpool   EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

WriteToSpoolFile Proc    Near
TestNearEOF:    Mov     SI, WriteChars
                Mov     DX, word ptr WritePosition+2
                Mov     AX, word ptr WritePosition
                Mov     CX, SI
                Call    IncPosition

                Cmp     DX, word ptr FileSize+2
                JA      JustToEOF
                JB      SeekForWrite
                Cmp     AX, word ptr FileSize
                JBE     SeekForWrite

JustToEOF:      Mov     DX, word ptr FileSize+2
                Mov     AX, word ptr FileSize
                Sub     AX, word ptr WritePosition
                Sbb     DX, word ptr WritePosition+2
                Mov     SI, AX

SeekForWrite:   Mov     CX, word ptr WritePosition+2
                Mov     DX, word ptr WritePosition
                Mov     AX, 4200h
                Mov     BX, SpoolFile
                PushF
                CLI
                Call    OldInt21

WriteIt:        Mov     AX, 4000h
                Mov     BX, SpoolFile
                Mov     CX, SI
                Mov     DX, Offset WriteBuffer
                PushF
                CLI
                Call    OldInt21
                JC      WriteError
                Cmp     AX, CX
                JNE     WriteError

WB_NewPos:      Mov     DX, word ptr WritePosition+2
                Mov     AX, word ptr WritePosition
                Call    IncPosition
                Mov     word ptr WritePosition+2, DX
                Mov     word ptr WritePosition, AX

                Sub     WriteChars, CX
                JNZ     TestNearEOF

WB_Exit:        Mov     WriteBufferFlag, 0
                Ret

WriteError:     Call    ErrorNoise
                Mov     BX, 10
                Call    Pause
                Call    ErrorNoise
                Mov     BX, 10
                Call    Pause
                Call    ErrorNoise
                Jmp     WB_Exit

WriteToSpoolFile EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

IncPosition     Proc    Near
                Add     AX, CX
                Adc     DX, 0

                Cmp     DX, word ptr FileSize+2
                JB      IP_Exit
                JA      ResetToStart
                Cmp     AX, word ptr FileSize
                JB      IP_Exit

ResetToStart:   Mov     DX, 0
                Mov     AX, 0

IP_Exit:        Ret
IncPosition     EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Beep            Proc    Near
                Push    AX
                Push    BX
                Push    CX

                STI
                Mov     AL, 0b6h
                Out     43h, AL

                Mov     BX, 0533h
                Mov     AL, BL
                Out     42h, AL
                Mov     AL, BH
                Out     42h, AL

                In      AL, 61h
                Mov     AH, AL
                Or      AL, 03h
                Out     61h, AL

                Mov     BX, 1
                Call    Pause

SpkrOff:        Mov     AL, AH
                Out     61h, AL

                Pop     CX
                Pop     BX
                Pop     AX
                Ret
Beep            EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Pause           Proc    Near
                CLI
                Mov     CX, word ptr TickCounter+2
                Mov     DX, word ptr TickCounter
                STI
                Add     DX, BX
                Adc     CX, 0

WaitBXTicks:    CLI
                Mov     AX, word ptr TickCounter+2
                Mov     BX, word ptr TickCounter
                STI

                Cmp     CX, AX
                JA      WaitBXTicks
                JB      P_Exit
                Cmp     DX, BX
                JA      WaitBXTicks

P_Exit:         Ret
Pause           EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

EndISR          db      0

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

;       this procedure converts an ASCII string into an integer (word)

;       input:  CX --- character count
;               BX --- address of string's 1st character
;       output: AX --- 2-byte integer
;               CF --- set if error, clear if okay

Ascii_Bin       Proc    Near
                Push    BX
                Push    CX
                Push    DX

                Sub     AX, AX
                Sub     DX, DX

                Cmp     CX, 0h
                JBE     AsciiBinExit

ScanWhiteSp:
                Cmp     Byte Ptr [BX], '-'
                JE      P0100
                Cmp     Byte Ptr [BX], '0'
                JB      ScanContinue
                Cmp     Byte Ptr [BX], '9'
                JA      ScanContinue
                Jmp     Short P0110
ScanContinue:
                Inc     BX
                Loop    ScanWhiteSp
                Jmp     Short AsciiBinError
P0100:
                Call    AsciiBinConvert
                JC      AsciiBinError
                Cmp     AX, 32768
                JA      AsciiBinError
                Neg     AX
                Jmp     Short AsciiBinExit
P0110:
                Call    AsciiBinConvert
                JC      AsciiBinError
                Cmp     AX, 32767
                JA      AsciiBinError
                Jmp     Short AsciiBinExit
AsciiBinError:
                STC
                Jmp     Short A_B_Exit
AsciiBinExit:
                CLC
A_B_Exit:
                Pop     DX
                Pop     CX
                Pop     BX

                Ret

Ascii_Bin       EndP


AsciiBinConvert Proc    Near
;  on entry, BX points to the string, and CX has the count
;  of the characters to be processed.  any leading white
;  space has been discarded.


                Push    BX
                Push    CX
                Push    DX
                Push    SI

                Sub     AX, AX
Chk_Dig:
                Mov     SI, 10
                Mul     SI
                Mov     DL, [BX]
                AND     DX, 000Fh
                Add     AX, DX
                JC      A_B_ConvExit

                Inc     BX
                Loop    Chk_Dig
A_B_ConvExit:
                Pop     SI
                Pop     DX
                Pop     CX
                Pop     BX

                Ret

AsciiBinConvert EndP



Bin_Ascii       Proc    Near
;  this procedure converts an integer (word) into an Ascii string.
;
;       input:  AX --- 2-byte integer
;               BX --- address of string to hold the result
;               CX --- length of string
;       output: BX --- points to processed string

                Push    BX
                Push    CX
                Push    DX
                Push    SI

                Cmp     CX, 0h
                JBE     No_More

                Push    AX
                Push    BX
Bin_Fill:
                Mov     Byte Ptr [BX], '0'
                Inc     BX
                Loop    Bin_Fill

                Pop     CX
                Mov     SI, 10
                OR      AX, AX
                JNS     Clr_Dvd
                Neg     AX
Clr_Dvd:
                Sub     DX, DX
                Div     SI
                Add     DX, '0'
                Dec     BX
                Mov     Byte Ptr [BX], DL

                Cmp     CX, BX
                JE      ChkSgn

                OR      AX, AX
                JNZ     Clr_Dvd

ChkSgn:
                Mov     BX, CX

                Pop     AX
                OR      AX, AX
                JNS     No_More

                Mov     Byte Ptr [BX], '-'

No_More:
                Pop     SI
                Pop     DX
                Pop     CX
                Pop     BX

                Ret

Bin_Ascii       EndP


;       This procedure converts a long integer (4 bytes, low byte 1st)
;       into a null-terminated ascii string.
;
;       inputs:
;               DX:AX --- the number to be converted.
;               DI    --- the address of the string area (> 30 bytes!)
;
;       outputs:
;               CX    --- length of string
;
;

LongToAscii     Proc    Near
                Push    AX
                Push    BX
                Push    DX
                Push    SI
                Push    DI

                Push    DS
                Pop     ES

                Mov     CX, DX
                MOV     BX, 10

                Test    CX, 8000h
                JZ      StartString

MoveMinus:      MOV     Byte Ptr [DI], "-"
                INC     DI
                NEG     CX
                NEG     AX
                SBB     CX, 0

StartString:    LEA     SI, TempString
                JCXZ    LTOA_40 

LTOA_30:        XCHG    AX, CX
                SUB     DX, DX
                DIV     BX
                XCHG    AX, CX
                DIV     BX
                MOV     Byte Ptr [SI], DL
                INC     SI
                JCXZ    LTOA_50
         
                JMP     Short LTOA_30

LTOA_40:        SUB     DX, DX
                DIV     BX
                MOV     Byte Ptr [SI], DL
                INC     SI

LTOA_50:        OR      AX, AX
                JNZ     LTOA_40

                LEA     CX, TempString
                NEG     CX
                ADD     CX, SI
                CLD

                Push    CX

LTOA_55:        DEC     SI
                MOV     AL, Byte Ptr [SI]
                SUB     AL, 0Ah
                JNB     LTOA_60

                ADD     AL, 3Ah
                JMP     Short LTOA_65

LTOA_60:        ADD     AL, 61h
         
LTOA_65:        STOSB
                LOOP    LTOA_55

                MOV     AL, 0
                STOSB

                Pop     CX

                Pop     DI
                Pop     SI
                Pop     DX
                Pop     BX
                Pop     AX

                RET
LongToAscii     EndP

;-----------------------------------------------------
;
; CONVINT: Convert string into longword
;
;       Inputs: SI = start address of string
;               CX = number of chars to convert
;
;       Output: AX = low byte of converted value
;               DX = high byte of converted value
;
;-----------------------------------------------------
;

CONVINT         PROC    NEAR
                PUSH    DI
                PUSH    SI
                PUSH    CX

                XOR     AX,AX
                MOV     LoVal,AX
                MOV     HiVal,AX
                MOV     Sign ,AX
;
; skip over leading spaces
;
CONV1:          CMP     Byte Ptr [SI],' '
                JNE     CONV2
                INC     SI
                DEC     CX
                JCXZ    CONV98          ;All spaces so exit
                JMP     SHORT CONV1

CONV98:         JMP     CONVEXIT

;
; CX = number of chars left to test; SI = pointer to first
;
CONV2:          CMP     Byte Ptr [SI],'-'
                JE      CONV3
                CALL    ISDIGIT
                JNE     CONV98
                JE      CONV4

CONV3:          XOR     AX,AX
                OR      AX,Sign
                JNE     CONV98
                MOV     AX,0FFFFh
                MOV     Sign,AX
                DEC     CX              ;Skip the minus sign
                INC     SI
                JCXZ    CONV98
                JMP     SHORT CONV1
;
; Count the number of successive digits for power of ten counter
;
CONV4:          XOR     BX,BX

CONV6:          CALL    ISDIGIT         ;Legal?
                JE      CONV7           ;Yes
                CMP     Byte Ptr [SI],' '
                JNE     CONV5
CONV7:          INC     SI
                INC     BX
                LOOP    CONV6

CONV5:          MOV     CX,BX
                JCXZ    CONV98

                MOV     DI,OFFSET DGROUP:IntTable
                XOR     BX,BX           ;PowerOfTen index

CONV9:          DEC     SI
                XOR     DX,DX           ;Zero DX
                MOV     DL,[SI]
                PUSH    CX
                CMP     DL,' '
                JE      CONV8           ;Skip imbedded spaces
                MOV     CX,DX
                AND     CX,000Fh
                JCXZ    CONV8A
CONVB:          CLC
                MOV     AX,[DI][BX]
                ADD     LoVal,AX
                MOV     AX,[DI+2][BX]
                ADC     HiVal,AX
                LOOP    CONVB
CONV8A:         ADD     BX,4

CONV8:          POP     CX
                LOOP    CONV9

CONVEXIT:       MOV     AX,LoVal
                MOV     DX,HiVal
                MOV     CX,Sign
                JCXZ    CONVEX
                OR      AX,AX           ;Avoid '-0'
                JNE     CONVE2
                OR      DX,DX
                JE      CONVEX

CONVE2:         NOT     AX
                NOT     DX
                ADD     AX,1
                ADC     DX,0

CONVEX:         POP     CX
                POP     SI
                POP     DI

                RET

CONVINT         ENDP

ISDIGIT         PROC    NEAR
                CMP     Byte Ptr [SI],'0'
                JB      ISDIG1
                CMP     Byte Ptr [SI],'9'
                JA      ISDIG1
                CMP     AX,AX
ISDIG1:         RET

ISDIGIT         ENDP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

GetToken        Proc    Near
                mov     BP, 0

NextChar:       jcxz    NoMoreStuff
                mov     al, byte ptr es:[bx]
                cmp     al, 0Dh
                je      NoMoreStuff
                cmp     al, 00h
                je      NoMoreStuff

Separators:     cmp     al, ","
                je      SkipChar
                cmp     al, " "
                je      SkipChar
                cmp     al, "/"
                je      SkipChar
                cmp     al, "-"
                je      SkipChar

ProcessChar:    mov     byte ptr [di], al
                inc     bp
                inc     di

BumpUp:         inc     bx
                dec     cx
                jmp     NextChar

SkipChar:       cmp     bp, 0
                je      BumpUp

NoMoreStuff:    mov     byte ptr [di], 0
                Ret
GetToken        EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

ParseCommandLine Proc   Near
                mov     es, OurPSP
                xor     cx, cx
                mov     cl, byte ptr es:[80h]

Defaults:       Mov     SpoolSize, 200
                Mov     Priority, 5
                Mov     DoFormFeeds, 0
                Mov     OutputName, 0

GetSpoolSize:   mov     bx, 81h
                mov     di, offset String1
                call    GetToken
                cmp     bp, 0
                je      ParseCmdExit

                push    cx
                push    bx
                Mov     cx, bp
                Mov     BX, Offset String1
                Call    Ascii_Bin
                pop     bx
                pop     cx
                Cmp     AX, 5
                JB      ShowUsage
                Cmp     AX, 999
                JA      ShowUsage
                Shl     AX, 1
                Mov     SpoolSize, AX

GetPriority:    mov     di, offset String1
                call    GetToken
                cmp     bp, 0
                je      ParseCmdExit

                push    cx
                push    bx
                Mov     cx, bp
                Mov     BX, Offset String1
                Call    Ascii_Bin
                pop     bx
                pop     cx
                Cmp     AX, 1
                JB      ShowUsage
                Cmp     AX, 20
                JA      ShowUsage
                Mov     Priority, AX

GetFFflag:      mov     di, offset String1
                call    GetToken
                cmp     bp, 0
                je      ParseCmdExit

                Cmp     String1, "Y"
                JE      SetFF
                Cmp     String1, "y"
                JE      SetFF
                Cmp     String1, "N"
                JE      GetOutputName
                Cmp     String1, "n"
                JE      GetOutputName
                Jmp     ShowUsage

SetFF:          Mov     DoFormFeeds, 1

GetOutputName:  Mov     di, offset OutputName
                call    GetToken

ParseCmdExit:   Ret

ShowUsage:      @Write  Msg8, LMsg8, StdErr
                @Exit   1

ParseCommandLine EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

CreateSpoolName Proc    Near
                Push    DS
                Pop     ES

                Mov     AH, 19h
                Int     21h
                Add     AL, "A"
                Mov     AH, ":"
                Mov     Word Ptr SpoolPathName, AX
                Mov     SpoolPathName+2, '\'

                Mov     SI, Offset SpoolPathName
                Inc     SI
                Inc     SI
                Inc     SI
                Mov     DL, 0
                Mov     AH, 47h
                Int     21h

                Mov     DI, Offset SpoolPathName
                Mov     CX, 128
                Mov     AL, 0
        RepNE   ScaSB
                Dec     DI
                Cmp     Byte Ptr [DI-1], '\'
                JE      DoNamePart
                Mov     Byte Ptr [DI], '\'
                Inc     DI

DoNamePart:     Mov     SI, Offset SpoolName
                Mov     CX, 10
        Rep     MovSB

                Ret
CreateSpoolName EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

ParseOutputName Proc    Near
                Push    DS
                Pop     ES

                Mov     AL, 0
                Mov     CX, 80
                Mov     DI, Offset FullName
        Rep     StoSB

                Mov     DI, Offset OutputName
                Mov     BX, DI
                Mov     CX, 80
                Mov     AL, 0
        RepNE   ScaSB
                Sub     DI, BX
                Mov     OutputLength, DI

                Mov     SI, Offset FullName

                Mov     DI, Offset OutputName
                Cmp     Byte Ptr [DI+1], ':'
                JE      Parse_Drive

                Mov     AH, 19h
                Int     21h
                Add     AL, "A"
                Mov     AH, ":"
                Jmp     Short Parse_GetName

Parse_Drive:    Mov     AX, Word Ptr OutputName
                Inc     DI
                Inc     DI
                Dec     OutputLength
                Dec     OutputLength
                Cmp     AL, '_'
                JBE     Parse_GetName
                And     AL, '_'

Parse_GetName:  Mov     Word Ptr FullName, AX
                Inc     SI
                Inc     SI

                Push    DI

                Cmp     Byte Ptr [DI], '\'      ; do we need cwd?
                JE      Parse_DoRemainder       ; no

                Mov     Byte Ptr [SI], '\'
                Inc     SI
                Mov     DL, Byte Ptr FullName
                Sub     DL, '@'
                Mov     AH, 47h
                Int     21h
                Mov     DI, SI
                Mov     CX, 80
                Mov     AL, 0
        RepNE   ScaSB
                Dec     DI
                Mov     SI, DI
                Cmp     Byte Ptr [SI-1], '\'
                JE      Parse_DoRemainder
                Mov     Byte Ptr [SI], '\'
                Inc     SI

Parse_DoRemainder:
                Mov     CX, OutputLength
                Mov     DI, SI
                Pop     SI
        Rep     MovSB
                Mov     Byte Ptr [DI], 0

                Mov     DI, Offset FullName
                Mov     BX, DI
                Mov     CX, 80
                Mov     AL, 0
        RepNE   ScaSB
                Dec     DI
                Sub     DI, BX
                Mov     OutputLength, DI
                Ret
ParseOutputName EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

InitProg        Proc    Near
                Mov     Toggle, 0

                @Write  Msg5, LMsg5, StdErr
                Mov     BP, 0

InitSpool:      @Write  WriteBuffer, 512, SpoolFile
                Cmp     AX, CX
                JNE     SplError

                Inc     BP
                Cmp     BP, SpoolSize
                JB      InitSpool

                Jmp     Short CloseIt

SplError:       @ClosFil SpoolFile
SplErrorMsg:    @Write  Msg4, LMsg4, StdErr
                @Exit   1

CloseIt:        @ClosFil SpoolFile
                Cmp     ShareFlag, 1
                JE      OpenDOS3

OpenDOS2:       @OpenFil SpoolPathName, 02h
                JC      SplErrorMsg
                Mov     SpoolFile, AX
                Jmp     SayCreateOk

OpenDOS3:       @OpenFil SpoolPathName, 22h
                JC      SplErrorMsg
                Mov     SpoolFile, AX

SayCreateOk:    @Write  Msg6, LMsg6, StdErr

                Mov     DX, 0
                Mov     AX, 512
                Mul     SpoolSize
                Mov     word ptr FileSize+2, DX
                Mov     word ptr FileSize, AX

                Mov     AX, 3400h
                Int     21h
                Mov     InDOS_Seg, ES
                Mov     InDOS_Ofs, BX

                Cmp     OSMajor, 2
                JNE     IP_20

                Inc     BX
                Jmp     Short IP_30

IP_20:          Dec     BX

IP_30:          Mov     InDOS2_Seg, ES
                Mov     InDOS2_Ofs, BX
                @GetDTA
                Mov     OurDTASeg, ES
                Mov     OurDTAOfs, BX

                Mov     AX, 5100h
                Int     21h
                Mov     OurPID, BX

                Mov     AX, OurPSP
                Dec     AX
                Mov     OurMCBSeg, AX
                Mov     OurMCBOfs, 0000
                Mov     OurMCBSizeSeg, AX
                Mov     OurMCBSizeOfs, 0003

                @GetInt 05h
                Mov     OldInt05Seg, ES
                Mov     OldInt05Ofs, BX
                @GetInt 08h
                Mov     OldInt08Seg, ES
                Mov     OldInt08Ofs, BX
                @GetInt 10h
                Mov     OldInt10Seg, ES
                Mov     OldInt10Ofs, BX
                @GetInt 13h
                Mov     OldInt13Seg, ES
                Mov     OldInt13Ofs, BX
                @GetInt 16h
                Mov     OldInt16Seg, ES
                Mov     OldInt16Ofs, BX
                @GetInt 17h
                Mov     OldInt17Seg, ES
                Mov     OldInt17Ofs, BX
                @GetInt 28h
                Mov     OldInt28Seg, ES
                Mov     OldInt28Ofs, BX
                @GetInt 20h
                Mov     OldInt20Seg, ES
                Mov     OldInt20Ofs, BX
                @GetInt 21h
                Mov     OldInt21Seg, ES
                Mov     OldInt21Ofs, BX

                Push    DS
                @SetInt 05h, Int_05
                @SetInt 10h, Int_10
                @SetInt 13h, Int_13
                @SetInt 16h, Int_16
                @SetInt 17h, Int_17
                @SetInt 20h, Int_20
                @SetInt 21h, Int_21
                @SetInt 28h, Int_28
                @SetInt 08h, Int_08
                Pop     DS

                Ret
InitProg        EndP

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

Start:          CLD
                Mov     AX, @Data
                Mov     DS, AX
                Mov     OurPSP, ES

                @GetMode
                Cmp     AL, 7
                JE      P10_Mono
                Mov     Low_Inten, 07h
                Mov     Hi_Inten, 0fh
                Jmp     Short P20_CheckInstall

P10_Mono:       Mov     Low_Inten, 07h
                Mov     Hi_Inten, 0fh

P20_CheckInstall:
                Mov     AX, "BN"
                Int     16h
                Cmp     AX, "bn"
                JE      AlreadyInstalled
                Jmp     NotPresent

AlreadyInstalled:
                @Write  Msg2, LMsg2, StdErr
                @Exit   0

NotPresent:     @GetVer
                Mov     OSMajor, AL
                Mov     OSMinor, AH
                Cmp     OSMajor, 2
                JAE     DosVerOkay

WrongDOS:       @Write  Msg3, LMsg3, StdErr
                @Exit   0

DosVerOkay:     Mov     ShareFlag, 0
                Cmp     OSMajor, 2
                JE      AfterShareTest

                Mov     AX, 1000h
                Int     2fh
                Cmp     AL, 0ffh
                JNE     AfterShareTest

                Mov     ShareFlag, 1

AfterShareTest: Call    CreateSpoolName

SaveStack:      Mov     save_ss, SS
                Mov     save_sp, SP

                Call    ParseCommandLine

                @MakFil SpoolPathName, 0
                JNC     SpoolCreated

                @Write  Msg4, LMsg4, StdErr
                @Exit   1

SpoolCreated:   Mov     SpoolFile, AX

                Cmp     OutputName, 0
                JE      InstallUs

CreateOutput:   Call    ParseOutputName
                @MakFil FullName, 0
                JNC     OutputCreated

SayOutputError: @Write  Msg7, LMsg7, StdErr
                @Exit   1

OutputCreated:  Mov     OutputFile, AX
                @ClosFil OutputFile
                Cmp     ShareFlag, 1
                JE      OpenOutDOS3

OpenOutDOS2:    @OpenFil FullName, 02h
                JC      SayOutputError
                Mov     OutputFile, AX
                Jmp     InstallUs

OpenOutDOS3:    @OpenFil FullName, 22h
                JC      SayOutputError
                Mov     OutputFile, AX

InstallUs:      Call    InitProg
                @Write  Msg1, LMsg1, StdErr

                Mov     AX, CS
                Sub     AX, OurPSP

                Mov     BX, Offset EndISR
                Mov     CX, 4
                Shr     BX, CL
                Inc     BX

                Add     AX, BX

                Mov     KeepSize, AX
                @TSR    KeepSize, 0

                End     Start

; - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

